import math
import uuid
import json
import time
import codecs
import shutil
import traceback
from pathlib import Path
from urllib.parse import urlparse
from collections import OrderedDict
from lyrebird import utils, application
from lyrebird.log import get_logger
from lyrebird.application import config
from lyrebird.mock import context
from lyrebird.mock.dm.match import MatchRules

PROP_FILE_NAME = '.lyrebird_prop'
logger = get_logger()


class DataManager:

    def __init__(self):
        self.root_path: Path = None
        self.display_data_map = {}
        self.id_map = {}
        self.activated_data = OrderedDict()
        self.activated_group = {}
        self.is_activated_data_rules_contains_request_data = False
        self.LEVEL_SUPER_ACTIVATED = 3
        self.clipboard = None
        self.save_to_group_id = None
        self.tmp_group = {'id': 'tmp_group', 'type': 'group', 'name': 'tmp-group', 'label': [], 'children': []}
        self.snapshot_import_cache = {}
        self.SNAPSHOT_SUFFIX = '.lb'
        self.COPY_NODE_NAME_SUFFIX = ' - copy'
        self.root = self.get_default_root()
        self._snapshot_workspace = None

        self.add_group_ignore_keys = set(['id', 'type', 'children'])
        self.update_group_ignore_keys = set(['id', 'parent_id', 'type', 'children'])
        self.supported_data_type = ['data', 'json']

        self.unsave_keys = set()

        self.DELETE_STEP = 100
        self.is_deleting_lock = False

    @property
    def snapshot_workspace(self):
        if not self._snapshot_workspace:
            self._snapshot_workspace = Path(application._cm.ROOT) / 'snapshot'
        if not self._snapshot_workspace.exists():
            self._snapshot_workspace.mkdir()
        return self._snapshot_workspace

    @snapshot_workspace.setter
    def snapshot_workspace(self, workspace):
        self._snapshot_workspace = workspace

    def get_default_root(self):
        return {
            'id': str(uuid.uuid4()),
            'name': '$',
            'type': 'group',
            'parent_id': None,
            'label': [],
            'children': []
        }

    def set_adapter(self, adapter_cls):
        # TODO: Load overridden function only
        self._adapter = adapter_cls(self)

    def set_root(self, uri):
        self._adapter._set_root(uri)

    def reload(self):
        self._adapter._reload()
        self.add_parent_into_node()
        self.add_super_by()

    def add_parent_into_node(self):
        self.add_parent(self.root)
        self.unsave_keys.update(['parent', 'abs_parent_path'])

    def add_parent(self, node, parent_node=None):
        parent_node_parent = parent_node['parent'] if parent_node and parent_node.get('parent') else []
        parent_obj = parent_node_parent + [{
            'id': node['id'],
            'name': node['name'],
            'type': node['type'],
            'parent_id': node['parent_id']
        }]

        abs_parent_path = f"{parent_node['abs_parent_path']}{node['name']}/" if parent_node else '/'

        node.update({
            'parent': parent_obj,
            'abs_parent_path': abs_parent_path
        })

        if not node.get('children'):
            return

        for child in node['children']:
            self.add_parent(child, node)

    def add_super_by(self):
        for node in self.id_map.values():
            if not node.get('super_id'):
                continue
            super_parent_node = self.id_map.get(node['super_id'])
            if not super_parent_node:
                continue
            if not super_parent_node.get('super_by'):
                super_parent_node['super_by'] = []
            super_parent_node['super_by'].append({
                'id': node['id'],
                'name': node['name']
            })
        self.unsave_keys.add('super_by')

    def get(self, _id):
        """
        Get mock group or data by id
        """
        if not self.root:
            raise RootNotSet
        node = self.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)
        if node.get('type') == 'group' or node.get('type') == None:
            return self._adapter._get_group(_id)
        elif node.get('type') == 'data':
            return self._adapter._load_data(_id)
        elif node.get('type') == 'json':
            return self._adapter._load_data(_id)
        else:
            raise UnsupportedType

    # -----
    # Mock operations
    # -----

    def make_data_map_by_group(self, group_ids):
        self.display_data_map = {k:v for k,v in self.root.items() if k!='children'}
        self.display_data_map['children'] = []

        for group_id in group_ids:
            map_pointer = self.display_data_map['children']

            group = self.id_map.get(group_id)
            group_abs_parent_obj = self._get_abs_parent_obj(group)

            # init parent
            # the last node in group['abs_parent_path'] is node itself, expand node itself
            # expand root which is already in the data_map
            for group_info in group_abs_parent_obj[1:-1]:
                parent_id = group_info['id']

                for index, child in enumerate(map_pointer):
                    if parent_id == child['id']:
                        map_pointer = map_pointer[index]['children']
                        break

                else:
                    node = self.id_map.get(parent_id)

                    new_node = {k:v for k,v in node.items() if k != 'children'}
                    new_node['children'] = []
                    map_pointer.append(new_node)
                    # Because of new_node is `append` into map_pointer, new_node is the last child of map_pointer
                    # move map_pointer first, sort after
                    map_pointer = map_pointer[-1]['children']
                    map_pointer.sort(key=lambda x:x.get('name', ''))

            # update node and node's children
            node = self.id_map.get(group_id)
            map_pointer.append(node)

        return self.display_data_map

    def activate(self, search_id, **kwargs):
        """
        Activite data by id
        """
        if not self.root:
            raise RootNotSet
        _node = self._adapter._get_activate_group(search_id)
        if not _node:
            raise IDNotFound(f'ID:{search_id}')

        _id = _node['id']
        node_id_list = self._collect_activate_node(_node, level_lefted=self.LEVEL_SUPER_ACTIVATED)

        ordered_data_id = []
        for node_id in node_id_list:
            activated_node = self.id_map.get(node_id)
            ordered_data_id += self._collect_data_in_node(activated_node)
        data_list = self._adapter._load_data_by_query({'id': ordered_data_id})

        # Only one group could be activated and reactived
        # Deactivate all activated group before activate
        self.deactivate()

        data_map = {d['id']: d for d in data_list}
        self.activated_data.update({i: data_map[i] for i in ordered_data_id if data_map.get(i)})
        self.activated_group[_id] = _node

        # After activate
        self._check_activated_data_rules_contains_request_data()
        self._adapter._after_activate(**kwargs)

    def _check_activated_data_rules_contains_request_data(self):
        self.is_activated_data_rules_contains_request_data = False
        for _data_id in self.activated_data:
            _data = self.activated_data[_data_id]
            rules = _data.get('rule')
            if not rules:
                continue
            for rule_key in rules.keys():
                if rule_key.startswith('request.data'):
                    self.is_activated_data_rules_contains_request_data = True
                    return 

    def _collect_activate_node(self, node, level_lefted=1):
        id_list = [node['id']]
        if level_lefted <= 0:
            return id_list
        if not node.get('super_id'):
            return id_list
        if node.get('super_id') == node['id']:
            raise SuperIdCannotBeNodeItself(node['id'])

        _super_id = node.get('super_id')
        _super_node = self.id_map.get(_super_id)
        if not _super_node:
            raise IDNotFound(f'Super node ID: {_super_id}')

        id_list.extend(self._collect_activate_node(_super_node, level_lefted=level_lefted-1))
        return id_list

    def _collect_data_in_node(self, node):
        id_list = []
        if node.get('type') in self.supported_data_type:
            return [node['id']]
        elif node.get('type', '') == 'group':
            if 'children' in node:
                for child in node['children']:
                    id_list.extend(self._collect_data_in_node(child))
        return id_list

    def deactivate(self):
        """
        Clear activated data
        """
        self.activated_data = OrderedDict()
        self.activated_group = {}
        self._check_activated_data_rules_contains_request_data()

    def reactive(self):
        for _group_id in self.activated_group:
            self.activate(_group_id)

    def get_matched_data(self, flow):
        """
        Find matched mock data from activated data
        """
        if self.is_activated_data_rules_contains_request_data:
            decode_flow = {}
            application.encoders_decoders.decoder_handler(flow, output=decode_flow)
        else:
            decode_flow = flow
        _matched_data = []
        for _data_id in self.activated_data:
            _data = self.activated_data[_data_id]
            if self._is_match_rule(decode_flow, _data.get('rule')):
                _matched_data.append(_data)
                break

        for response_data in _matched_data:
            if 'response' not in response_data:
                continue
            if 'data' not in response_data['response']:
                continue
            if not response_data['response']['data']:
                continue
            self._format_respose_data(response_data)

        return _matched_data

    @staticmethod
    def _format_respose_data(flow):
        # TODO render mock data before response, support more functions
        origin_response_data = flow['response']['data']

        try:
            flow['response']['data'] = utils.render(flow['response']['data'])
        except Exception:
            flow['response']['data'] = origin_response_data
            logger.warning(f'Format response data error! {flow["request"]["url"]}\n {traceback.format_exc()}')

    def _is_match_rule(self, flow, rules):
        return MatchRules.match(flow, rules)

    # -----
    # Data tree operations
    # -----

    def _get_request_path(self, request):
        path = request.get('path')
        if not path:
            if not request.get('url'):
                return ''
            parsed_url = urlparse(request['url'])
            path = parsed_url.path
        return path

    def _make_data(self, raw_data, **kwargs):
        data = dict(raw_data)
        _data_id = kwargs.get('data_id') or str(uuid.uuid4())
        data['id'] = _data_id

        if data.get('type') == 'json':
            self._make_data_json(raw_data, data)
        elif data.get('type') == 'data':
            self._make_data_http_data(raw_data, data)
        else:
            self._make_data_http_data(raw_data, data)

        return data

    def _make_data_http_data(self, raw_data, data):
        if 'request' in data:
            # TODO remove it with inspector frontend
            data['request'] = dict(raw_data['request'])

            _data_name = data['name'] if data.get('name') else self._adapter._get_data_name(data)
            _data_rule = data['rule'] if data.get('rule') else self._adapter._get_data_rule(data['request'])
            if 'data' in data['request']:
                data['request']['data'] = self._flow_data_2_str(data['request']['data'])
        else:
            _data_name = data.get('name')
            _data_rule = {'request.url': '(?=.*YOUR-REQUEST-PATH)(?=.*PARAMS)'}
            data['request'] = {}

        if 'response' in data:
            # TODO remove it with inspector frontend
            data['response'] = dict(raw_data['response'])

            if 'data' in data['response']:
                data['response']['data'] = self._flow_data_2_str(data['response']['data'])
        else:
            data['response'] = {}

        # proxy_response will not be saved
        if 'proxy_response' in data:
            del data['proxy_response']

        data['name'] = _data_name
        data['rule'] = _data_rule

    def _make_data_json(self, raw_data, data):
        if 'json' not in data:
            data['json'] = {}

        data['name'] = data['name'] if data.get('name') else 'Unnamed JSON'

    def add_data(self, parent_id, raw_data, **kwargs):
        if not isinstance(raw_data, dict):
            raise DataObjectShouldBeADict

        if not raw_data.get('type'):
            raw_data['type'] = 'data'

        if raw_data['type'] not in self.supported_data_type:
            raise UnsupportedType

        parent_node = None
        if parent_id == 'tmp_group':
            parent_node = self.tmp_group
        elif parent_id:
            parent_node = self.id_map.get(parent_id)
            if not parent_node:
                raise IDNotFound(parent_id)
            if parent_node['type'] != 'group':
                raise OnlyGroupCanContainAnyOtherObject

        data = self._make_data(raw_data)
        data['parent_id'] = parent_id
        _data_id = data['id']
        _data_name = data.get('name')

        output_path = (kwargs['output'] / _data_id) if kwargs.get('output') else None

        self._adapter._add_data(data, path=output_path)

        if parent_node:
            data_node = {}
            data_node['id'] = _data_id
            data_node['name'] = _data_name
            data_node['type'] = data.get('type')
            data_node['parent_id'] = parent_id

            # New data added in the head of child list
            parent_node['children'].insert(0, data_node)
            logger.debug(f'*** Add to node {data_node}')
            # Update ID mapping
            self.id_map[_data_id] = data_node
            self._adapter._add_group(data_node)

        return _data_id

    def _flow_data_2_str(self, data):
        if isinstance(data, str):
            return data
        return json.dumps(data, ensure_ascii=False)

    def add_group(self, data):
        ignore_key = self.add_group_ignore_keys | self.unsave_keys
        new_group = {k: data[k] for k in data if k not in ignore_key}

        parent_id = data.get('parent_id')
        if parent_id == None:
            parent_node = self.root
        else:
            parent_node = self.id_map.get(parent_id)

        if not parent_node:
            raise IDNotFound(parent_id)
        if parent_node.get('type') == 'data':
            raise OnlyGroupCanContainAnyOtherObject

        # Add group
        group_id = str(uuid.uuid4())
        new_group.update({
            'id': group_id,
            'type': 'group',
            'children': []
        })
        # New group added in the head of child list
        if 'children' not in parent_node:
            parent_node['children'] = []
        parent_node['children'].insert(0, new_group)
        # Register ID
        self.id_map[group_id] = new_group
        # Save prop
        self._adapter._add_group(new_group)
        return group_id

    def add_group_by_path(self, path):
        parent_name = path.strip('/')
        current = self.root
        if not parent_name:
            return current['id']
        parent_name_list = parent_name.split('/')
        for name in parent_name_list:
            for child in current['children']:
                if name == child['name']:
                    current = child
                    break
            else:
                group_id = self.add_group({
                    'name': name,
                    'parent_id': current['id']
                })
                current = self.id_map.get(group_id)
        return current['id']

    def delete(self, _id):
        target_node = self.id_map.get(_id)
        if not target_node:
            raise IDNotFound(_id)
        parent_id = target_node.get('parent_id')
        # Remove refs
        if parent_id:
            parent = self.id_map.get(parent_id)
            parent['children'].remove(target_node)
        else:
            self.root['children'].remove(target_node)
        self._delete(_id)

    def _delete(self, _id):
        target_node = self.id_map.get(_id)
        if not target_node:
            raise IDNotFound(_id)
        if 'children' in target_node and len(target_node['children']) > 0:
            for child in target_node['children'][::-1]:
                self._delete(child['id'])
                target_node['children'].pop(-1)
        # Remove from activated_group
        if _id in self.activated_group:
            self.activated_group.pop(_id)
        # Delete from ID mapping
        self.id_map.pop(_id)
        # Delete from mock tree
        self._adapter._delete_group(_id)
        # Delete from file system
        if target_node['type'] == 'data':
            self._adapter._delete_data(_id)

    def cut(self, _id):
        _node = self.id_map.get(_id)
        if not _node:
            raise IDNotFound(_id)
        self.clipboard = {
            'type': 'cut',
            'id': _id,
            'node': _node
        }

    def copy(self, _id):
        _node = self.id_map.get(_id)
        if not _node:
            raise IDNotFound(_id)
        self.clipboard = {
            'type': 'copy',
            'id': _id,
            'node': _node
        }

    def import_(self, node, json=None, path=None):
        self.clipboard = {
            'type': 'import',
            'id': node['id'],
            'node': node,
            'json': json,
            'path': path
        }

    def paste(self, parent_id, **kwargs):
        if not self.clipboard:
            raise NoneClipbord
        _parent_node = self.id_map.get(parent_id)
        _node = self.clipboard['node']
        if not _parent_node:
            raise IDNotFound(parent_id)
        if not _parent_node.get('children'):
            _parent_node['children'] = []
        if self.clipboard['type'] == 'cut':
            _origin_parent = self.id_map.get(_node['parent_id'])
            _origin_parent['children'].remove(_node)
            _parent_node['children'].insert(0, _node)
            _node['parent_id'] = parent_id
            _group_id = _node['id']
            self._adapter._update_group(_node)
        elif self.clipboard['type'] == 'copy':
            new_name = self._get_copy_node_new_name(_node)
            _group_id = self._copy_node(_parent_node, _node, name=new_name, origin_name=_node['name'], **kwargs)
        elif self.clipboard['type'] == 'import':
            _group_id = self._copy_node(_parent_node, _node, **kwargs)
        return _group_id

    def duplicate(self, _id, **kwargs):
        return self._adapter.duplicate(_id)

    def _copy_node(self, parent_node, node, **kwargs):
        new_node = {}
        new_node.update(node)
        new_node['id'] = str(uuid.uuid4())
        new_node['parent_id'] = parent_node['id']
        new_node['name'] = kwargs['name'] if kwargs.get('name') else new_node['name']
        # Add to target node
        if not parent_node.get('children'):
            parent_node['children'] = []
        parent_node['children'].insert(0, new_node)
        # Add node
        self._adapter._add_group(new_node, **kwargs)
        # Register ID
        self.id_map[new_node['id']] = new_node
        if new_node['type'] == 'group':
            kwargs.pop('name') if kwargs.get('name') else None
            kwargs.pop('origin_name') if kwargs.get('origin_name') else None
            new_node['children'] = []
            for child in node['children']:
                self._copy_node(new_node, child, **kwargs)
        elif new_node['type'] == 'data':
            self._copy_data(new_node, node, **kwargs)
        elif new_node['type'] == 'json':
            self._copy_data(new_node, node, **kwargs)
        return new_node['id']

    def _copy_data(self, new_data_node, data_node, **kwargs):
        _id = data_node['id']

        if self.clipboard.get('json'):
            event = self.clipboard.get('json', {}).get(_id)
            if not event:
                raise DataNotFound
            prop = {k:v for k,v in event.items()}
        elif self.clipboard.get('path'):
            with codecs.open(Path(self.clipboard['path'])/_id) as f:
                prop = json.load(f)
        else:
            prop = self._adapter._load_data(_id)

        prop['id'] = new_data_node['id']
        prop['name'] = kwargs.pop('name') if kwargs.get('name') else prop['name']
        self._adapter._add_data(prop)

    def _get_copy_node_new_name(self, _node):
        return _node['name'] + self.COPY_NODE_NAME_SUFFIX

    def _sort_children_by_name(self):
        for node_id in self.id_map:
            node = self.id_map[node_id]
            if 'children' not in node:
                # fix mock data group has no children
                if node['type'] == 'group':
                    node['children'] = []
                continue
            node['children'] = sorted(node['children'], key=lambda sub_node: sub_node['name'])

    # -----
    # Conflict checker
    # -----

    def check_conflict(self, _id):
        node = self.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)
        data_array = []

        def _read_data(node):
            if node.get('type') == 'data':
                _data = self._adapter._load_data(node['id'])
                _data['parent_id'] = node['parent_id']
                data_array.append(_data)
            elif node.get('type') == 'group':
                for child in node['children']:
                    _read_data(child)
        _read_data(node)
        return self.check_conflict_data(data_array)

    def activated_data_check_conflict(self):
        data_array = list(self.activated_data.values())
        return self.check_conflict_data(data_array)

    def check_conflict_data(self, data_array):
        conflict_rules = []
        for _data in data_array:
            _rule = _data['rule']
            _hit_data = []
            for _test_data in data_array:
                if self._is_match_rule(_test_data, _rule):
                    _target_node = {
                        'id': _test_data['id'],
                        'name': _test_data['name'],
                        'url': _test_data['request']['url'],
                        'abs_parent_path': self._get_abs_parent_path(_test_data)
                    }
                    _hit_data.append(_target_node)
            if len(_hit_data) > 1:
                _src_node = {
                    'id': _data['id'],
                    'name': _data['name'],
                    'rule': _data['rule'],
                    'abs_parent_path': self._get_abs_parent_path(_data)
                }
                conflict_rules.append(
                    {
                        'data': _src_node,
                        'conflict_data': _hit_data
                    }
                )
        return conflict_rules

    def _get_abs_parent_path(self, node, path='/'):
        parent_node = self._get_node_parent(node)
        if parent_node is None:
            return path
        current_path = '/' + node['name'] + path
        return self._get_abs_parent_path(parent_node, path=current_path)

    def _get_abs_parent_obj(self, node, parent_obj=None):
        if parent_obj is None:
            parent_obj = []
        if 'id' not in node:
            return parent_obj
        node_info = self.id_map.get(node['id'])
        if not node_info:
            return parent_obj
        parent_obj.insert(0, {
            'id': node_info['id'],
            'name': node_info['name'],
            'type': node_info['type'],
            'parent_id': node_info['parent_id']
        })
        parent_node = self._get_node_parent(node)
        if parent_node is None:
            return parent_obj
        return self._get_abs_parent_obj(parent_node, parent_obj=parent_obj)

    def _get_node_parent(self, node):
        if 'parent_id' not in node:
            return None
        parent_node = self.id_map.get(node['parent_id'])
        if not parent_node:
            return None
        return parent_node

    # -----
    # Record API
    # -----

    def save_data(self, data):
        if len(self.activated_group) > 0:
            # TODO use self.save_to_group_id
            target_group_id = list(self.activated_group.keys())[0]
            self.add_data(target_group_id, data)
        else:
            self.add_data('tmp_group', data)

    # -----
    # Editor
    # -----

    def update_group(self, _id, data, save=True):
        node = self.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)

        # Remove unsave_key add by DataManager itself
        for key in self.unsave_keys:
            data.pop(key) if key in data else None

        # Uneditable key
        for key in self.update_group_ignore_keys:
            data[key] = node.get(key)

        # labels are ordered, sorted by label name
        if 'label' not in data:
            data['label'] = []
        elif 'label' in data and isinstance(data['label'], list):
            data['label'].sort(key=lambda x:x.get('name', '').lower())

        message = self._adapter._update_group(data) if save else None

        # Update node
        # 1. Add new key into node
        update_data = {k: data[k] for k in data if k not in self.update_group_ignore_keys}
        node.update(update_data)

        # 2. Remove deleted key in node
        delete_keys = [k for k in node if k not in data and k not in self.update_group_ignore_keys]
        for key in delete_keys:
            node.pop(key)

        # 3. Update existed value
        for key, value in data.items():
            if key in node:
                node[key] = value

        return message

    def update_data(self, _id, data):
        node = self.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)
        node['name'] = data['name']
        self._adapter._update_data(data)
        self._adapter._update_group(node)

    # -----
    # Snapshot
    # -----

    def export_from_local(self, event):
        group_id = self.import_from_local(event)
        filename = self.export_from_remote(group_id)
        return group_id, filename

    def export_from_remote(self, node_id):
        snapshot_path = self._get_snapshot_path()
        node = self.id_map.get(node_id)
        self._write_file(snapshot_path/PROP_FILE_NAME, node)

        ordered_data_id = self._collect_data_in_node(node)
        data_list = self._adapter._load_data_by_query({'id': ordered_data_id})
        for mock_data in data_list:
            self._write_file(snapshot_path/mock_data['id'], mock_data)

        filename = utils.compress_tar(snapshot_path, snapshot_path, suffix=self.SNAPSHOT_SUFFIX)
        self._remove_file([snapshot_path])
        return filename

    def import_from_local(self, event):
        _prop = event['snapshot']
        parent_path = config.get('snapshot.import.workspace', '/')
        parent_id = self.add_group_by_path(parent_path)

        new_data_map = {}
        for e in event['events']:
            data = self._make_data(e, data_id=e['id'])
            new_data_map[e['id']] = data

        self.import_(_prop, json=new_data_map)
        return self.paste(parent_id)

    def import_from_file(self, parent_id, input_path, **kwargs):
        snapshot_info, output_path = self.get_snapshot_file_detail(input_path)
        if kwargs.get('name'):
            snapshot_info['name'] = kwargs['name']

        self.import_(snapshot_info, path=output_path)
        _group_id = self.paste(parent_id=parent_id, path=output_path)
        self._remove_file([input_path, output_path])
        return _group_id

    def read_snapshot_from_link(self, link):
        snapshot_path = self._get_snapshot_path()
        snapshot_filename = Path(f'{snapshot_path}{self.SNAPSHOT_SUFFIX}')

        utils.download(link, snapshot_filename)
        return snapshot_filename

    def get_snapshot_file_detail(self, input_path):
        try:
            output_path = utils.decompress_tar(input_path)
        except Exception as e:
            raise LyrebirdSnapshotBroken(e)

        snapshot_prop = output_path / PROP_FILE_NAME

        if not snapshot_prop.exists():
            raise LyrebirdPropNotExists
        with codecs.open(str(snapshot_prop)) as f:
            snapshot_info = json.load(f)
        return snapshot_info, output_path

    def _write_file(self, path, data):
        data_str = json.dumps(data, ensure_ascii=False)
        with codecs.open(path, 'w') as f:
            f.write(data_str)

    def _remove_file(self, files):
        for filepath in files:
            path = Path(filepath)
            if path.is_dir() and path.exists():
                shutil.rmtree(path)
            elif path.is_file() and path.exists():
                path.unlink()

    def _get_snapshot_path(self):
        temp_dir_name = f"{time.strftime('%Y-%m-%d-%H-%M-%S', time.localtime())}-{str(uuid.uuid4())}"
        snapshot_path = self.snapshot_workspace / temp_dir_name
        snapshot_path.mkdir()
        return snapshot_path

    def _get_type_hashmap(self, ids):
        type_map = {
            'group': [],
            'data': [],
            'json': []
        }
        for id_ in ids:
            node = self.id_map.get(id_)
            if not node:
                continue
            node_type = node['type']
            if node_type not in type_map:
                continue
            type_map[node_type].append(id_)
        return type_map

    def _delete_remove_from_parent(self, id_):
        target_node = self.id_map.get(id_)
        if not target_node:
            return
        parent_id = target_node.get('parent_id')
        if not parent_id:
            return
        parent = self.id_map.get(parent_id)
        if not parent:
            return
        parent['children'].remove(target_node)


    # batch action
    def delete_by_query(self, query):
        all_id_list = query.get('id') or []

        times = math.ceil(len(all_id_list) / self.DELETE_STEP)
        for index in range(times):
            start_point = index * self.DELETE_STEP
            stop_point = (index + 1) * self.DELETE_STEP
            id_list = all_id_list[start_point : stop_point]

            context.application.socket_io.emit('statusBarProcess', {
                'message': f'DataManager deleting ({start_point}/{len(all_id_list)})',
                'state': 'process'
            })

            type_map = self._get_type_hashmap(id_list)
            node_delete_group_ids = type_map['group'] + type_map['data'] + type_map['json']
            node_delete_data_ids = type_map['data'] + type_map['json']

            for id_ in id_list:
                # Delete from parent
                self._delete_remove_from_parent(id_)

                # Remove from activated_group
                if id_ in self.activated_group:
                    self.activated_group.pop(id_)

                # Delete from ID mapping
                self.id_map.pop(id_)

            # Delete
            self._adapter._delete_group_by_query({'id': node_delete_group_ids})
            self._adapter._delete_data_by_query({'id': node_delete_data_ids})

        context.application.socket_io.emit('statusBarProcess', {
            'message': f'DataManager delete finish! ({len(all_id_list)}/{len(all_id_list)})',
            'state': 'finish'
        })

        context.emit('datamanagerUpdate')


# -----------------
# Exceptions
# -----------------

class DumpPropError(Exception):
    pass



class RootNotSet(Exception):
    pass


class RootPathNotExists(Exception):
    pass


class RootPathIsNotDir(Exception):
    pass


class LyrebirdPropNotExists(Exception):
    pass


class DataNotFound(Exception):
    pass


class OnlyGroupCanContainAnyOtherObject(Exception):
    pass


class UnsupportedType(Exception):
    pass


class SuperIdCannotBeNodeItself(Exception):
    pass


class DataObjectShouldBeADict(Exception):
    pass


class IDNotFound(Exception):
    pass


class NoneClipbord(Exception):
    pass


class NonePropFile(Exception):
    pass


class TooMuchPropFile(Exception):
    pass


class NodeExist(Exception):
    pass


class LyrebirdSnapshotBroken(Exception):
    pass



import json
from lyrebird.log import get_logger

logger = get_logger()

class MockDataHelper:

    @staticmethod
    def read_mock_data(handler_context):
        content_type = handler_context.flow['response']['headers'].get('Content-Type', '')

        try:
            if content_type.startswith('application/json'):
                handler_context.flow['response']['data'] = json.loads(handler_context.flow['response']['data'])
            elif content_type.startswith('application/x-www-form-urlencoded'):
                handler_context.flow['response']['data'] = json.loads(handler_context.flow['response']['data'])

        except Exception as e:
            logger.warning(f'Convert mock data response failed. {e}')



from .form import FormHandler
from .json import JsonHandler
from .text import TextHandler
from .javascript import JavascriptHandler
from .default import DefaultHandler
from lyrebird.log import get_logger

logger = get_logger()

content_type_handlers = {
    'application/x-www-form-urlencoded': FormHandler,
    'application/javascript': JavascriptHandler,
    'application/json': JsonHandler,
    'text/': TextHandler
}

def _get_matched_action(content_type):
    for pattern, func in content_type_handlers.items():
        if content_type.startswith(pattern):
            return func
    return DefaultHandler

def origin2flow(content_type, request_data, chain=None):
    func = _get_matched_action(content_type)
    try:
        _data = func.origin2flow(request_data)
    except Exception as e:
        func = DefaultHandler
        _data = func.origin2flow(request_data)
        logger.warning(f'Convert Content-Type: {content_type} data origin2flow failed! {e}')
    finally:
        chain.append(func)

    return _data

def flow2origin(content_type, flow_data):
    func = _get_matched_action(content_type)
    try:
        _data = func.flow2origin(flow_data)
    except Exception as e:
        _data = DefaultHandler.flow2origin(flow_data)
        logger.warning(f'Convert Content-Type: {content_type} data flow2origin failed! {e}')
    return _data



from .manager import main, run
from .mock import context
from .mock.dm.jsonpath import jsonpath
from .mock.handlers.handler_context import HandlerContext
import os
from .event import CustomEventReceiver
from .checker import event
from .checker import encoder, decoder
from .checker import on_request, on_response, on_request_upstream, on_response_upstream
from .plugins import get_plugin_storage
from lyrebird import application
from lyrebird.log import get_logger

APPLICATION_CONF_DIR = os.path.join(os.path.expanduser('~'), '.lyrebird')


def start_background_task(target, *args, **kwargs):
    """Start a background task in a new thread

    :param target: task function
    :param args: args will be pass to the task function
    :param kwargs: kwargs will be pass to the task function
    :return: null
    """
    context.application.socket_io.start_background_task(target, *args, **kwargs)


def emit(event, *args, **kwargs):
    """
    Send socketio event

    """
    context.application.socket_io.emit(event, *args, **kwargs)


def subscribe(channel, func, *args, **kwargs):
    """
    订阅信号

    :param signal: 信号，包含lyrebird, overbridge, android, tracking, perf, hunter
    :param func: 响应事件的回调函数
    :param sender: 信号发送者标识
    """
    # context.application.event_bus.subscribe(channel, func)
    application.server['event'].subscribe(channel, func, *args, **kwargs)


def publish(channel, event, *args, **kwargs):
    """
    发送信号

    :param signal: 信号，包含lyrebird, overbridge, android, tracking, perf, hunter
    :param sender: 信号发送者标识
    :param kwargs: 事件数据
    :return:
    """
    # context.application.event_bus.publish(channel, event)
    application.server['event'].publish(channel, event, *args, **kwargs)


def add_background_task(name, func):
    """
    添加后台任务

    """
    application.server['task'].add_task(name, func)



"""
获取 Event server state
"""


class StateProxy:

    def get(self, index):
        return application.server['event'].state.get(index)

    def __getitem__(self, index):
        return application.server['event'].state[index]


state = StateProxy()


def report(data):
    application.reporter.report(data)



import json
import math
import datetime
import traceback
import time
from queue import Queue
from pathlib import Path
from lyrebird import application
from lyrebird import log
from lyrebird.utils import JSONFormat
from lyrebird.base_server import ThreadServer
from lyrebird.mock import context
from sqlalchemy import event, or_
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, Integer, Text, DateTime, create_engine, Table, MetaData
from sqlalchemy.orm.attributes import InstrumentedAttribute


"""
Database server

Worked as a backgrund thread
storage all event message into database
"""

logger = log.get_logger()

Base = declarative_base()


class LyrebirdDatabaseServer(ThreadServer):
    def __init__(self, path=None):
        self.database_uri = None
        super().__init__()

        if not path or path.isspace():
            ROOT_DIR = application.root_dir()
            DEFAULT_NAME = 'lyrebird.db'
            self.database_uri = ROOT_DIR/DEFAULT_NAME
        else:
            self.database_uri = Path(path).expanduser().absolute()

        self.init_engine()

        # init queue
        self.storage_queue = Queue()

        # subscribe all channel
        application.server['event'].subscribe('any', self.event_receiver)
    
    def auto_alter_tables(self, engine):
        metadata = MetaData()
        tables = {
            table_name: {
                column.name: column for column in Table(table_name, metadata, autoload=True, autoload_with=engine).c
            }
            for table_name in engine.table_names()
        }
        for model_class in Base.__subclasses__():
            table_name = model_class.__table__.name
            if table_name not in tables:
                continue
            table = tables[table_name]
            for attr_name in dir(model_class):
                try:
                    attr = getattr(model_class, attr_name)
                except:
                    logger.warning(f'[Local DB]Cannot get attr:{attr_name} from {model_class}')
                    continue
                if not isinstance(attr, InstrumentedAttribute):
                    continue
                if not hasattr(attr, 'type'):
                    continue
                if not hasattr(attr, 'compile'):
                    continue
                attr_name = attr.name 
                if attr_name in table:
                    continue
                column_type = attr.type.compile(dialect=engine.dialect)
                engine.execute(f'ALTER TABLE {table_name} ADD COLUMN {attr_name} {column_type}')

    def init_engine(self):
        sqlite_path = 'sqlite:///'+str(self.database_uri)+'?check_same_thread=False'

        engine = create_engine(str(sqlite_path))

        # Set pragma on connect
        # https://www.sqlite.org/pragma.html
        event.listen(engine, 'connect', self._fk_pragma_on_connect)

        # Create all tables
        Base.metadata.create_all(engine)
        # Create session factory
        session_factory = sessionmaker(bind=engine)
        Session = scoped_session(session_factory)
        self._scoped_session = Session
        self.auto_alter_tables(engine=engine)

        logger.info(f'Init DB engine: {self.database_uri}')

    def _fk_pragma_on_connect(self, dbapi_con, con_record):
        # https://www.sqlite.org/pragma.html#pragma_journal_mode
        dbapi_con.execute('PRAGMA journal_mode=MEMORY')
        # https://www.sqlite.org/pragma.html#pragma_synchronous
        dbapi_con.execute('PRAGMA synchronous=OFF')

    def event_receiver(self, event, channel=None, event_id=None):
        # event is decoded , which should be encoded when save
        # event is deepcopy when created, no needs to copy again
        if channel == 'flow':
            application.encoders_decoders.encoder_handler(event['flow'])

        content = json.dumps(event, ensure_ascii=False)
        if isinstance(event, dict):
            message = event.get('message')
        else:
            message = None
        flow = Event(event_id=event_id, channel=channel, content=content, message=message)
        self.storage_queue.put(flow)

    def run(self):
        session = self._scoped_session()
        while self.running:
            try:
                event = self.storage_queue.get()
                session.add(event)
                session.commit()
                context.emit('db_action', 'add event log')
            except Exception:
                logger.error(f'Save event failed. {traceback.format_exc()}')

    def stop(self):
        super().stop()

    @property
    def session(self):
        return self._scoped_session()

    def get_event(self, channel_rules, offset=0, limit=20, search_str=''):
        search_str_list = [item.strip() for item in search_str.strip().split('+')] if search_str else []
        session = self._scoped_session()
        _subquery = session.query(Event.id).order_by(Event.id.desc())
        if len(channel_rules) > 0:
            _subquery = _subquery.filter(Event.channel.in_(channel_rules))
        if len(search_str_list) > 0:
            _subquery = _subquery.filter(Event.message != None)
            or_cond = []
            for search_str in search_str_list:
                or_cond.append(Event.message.like(f'%%{search_str}%%'))
            _subquery = _subquery.filter(or_(*or_cond))
        _subquery = _subquery.offset(offset).limit(limit).subquery()
        result = session.query(Event).filter(Event.id == _subquery.c.id).all()
        self._scoped_session.remove()
        return result

    def get_page_index_by_event_id(self, event_id, channel_rules, limit=20):

        session = self._scoped_session()
        _subquery = session.query(Event).filter(Event.event_id == event_id).subquery()
        query = session.query(Event.id)
        if len(channel_rules) > 0:
            query = query.filter(Event.channel.in_(channel_rules))
        # Fix bug: event_id in page end, return next page
        # Wrong code: query.filter(Event.id>=subquery.c.id).count()
        result = query.filter(Event.id > _subquery.c.id).count()
        return int(result/limit)

    def get_channel_list(self):
        session = self._scoped_session()
        result = session.query(Event.channel) \
            .group_by(Event.channel) \
            .all()
        self._scoped_session.remove()
        return result

    def get_page_count(self, channel_rules, page_size=20, search_str=''):
        session = self._scoped_session()
        query = session.query(Event.id)
        search_str_list = [item.strip() for item in search_str.strip().split('+')] if search_str else []
        if len(channel_rules) > 0:
            query = query.filter(Event.channel.in_(channel_rules))
        if len(search_str_list) > 0:
            query = query.filter(Event.message != None)
            or_cond = []
            for search_str in search_str_list:
                or_cond.append(Event.message.like(f'%%{search_str}%%'))
            query = query.filter(or_(*or_cond))
        result = query.count()
        self._scoped_session.remove()
        return math.ceil(result / page_size)

    def reset(self):
        self.stop()
        self.database_uri.unlink()
        self.init_engine()
        # TODO After self.stop() could terminate Thread, change `self.running = True` into self.start()
        self.running = True


class Event(Base, JSONFormat):
    __tablename__ = 'event'

    id = Column(Integer, primary_key=True, autoincrement=True)
    channel = Column(String(16), index=True)
    event_id = Column(String(32), index=True)
    content = Column(Text)
    message = Column(Text, default=None)
    _timestamp = Column('timestamp', DateTime(timezone=True), default=datetime.datetime.utcnow)

    @hybrid_property
    def timestamp(self):
        seconds_offset = time.localtime().tm_gmtoff
        return self._timestamp.timestamp() + seconds_offset



import re
import traceback
from lyrebird.log import get_logger

logger = get_logger()


class FunctionExecutor:

    @staticmethod
    def func_handler(func_list, flow, handler_type='flow_editor'):
        for func_info in func_list:
            handler_fn = func_info['func']
            try:
                handler_fn(flow)
                # TODO: The flow is changed or not?
                action = {
                    'id': handler_type,
                    'name': func_info['name']
                }
                if flow.get('action'):
                    flow['action'].append(action)
                else:
                    flow['action'] = [action]
            except Exception:
                logger.error(traceback.format_exc())

    @staticmethod
    def get_matched_sorted_handler(func_list, flow):
        matched_func = []
        for func in func_list:
            rules = func['rules']
            if not rules or FunctionExecutor._is_req_match_rule(rules, flow):
                matched_func.append(func)
        matched_sorted_func = sorted(matched_func, key = lambda f:f['rank'])
        return matched_sorted_func

    @staticmethod
    def _is_req_match_rule(rules, flow):
        for rule_key in rules:
            pattern = rules[rule_key]
            target = FunctionExecutor._get_rule_target(rule_key, flow)
            if not target or not re.search(pattern, target):
                return False
        return True

    @staticmethod
    def _get_rule_target(rule_key, flow):
        prop_keys = rule_key.split('.')
        result = flow
        for prop_key in prop_keys:
            result = result.get(prop_key)
            if not result:
                return None
        return result



import logging
from logging.handlers import TimedRotatingFileHandler
from colorama import Fore, Style, Back
from collections import namedtuple
from pathlib import Path

DEFAULT_LOG_PATH = '~/.lyrebird/lyrebird.log'

LOGGER_INITED = False


Color = namedtuple('Color', ['fore', 'style', 'back'])

COLORS = dict(
    NOTICE=Color(fore=Fore.GREEN, style=Style.NORMAL, back=Back.RESET),
    CRITICAL=Color(fore=Fore.WHITE, style=Style.BRIGHT, back=Back.RED),
    ERROR=Color(fore=Fore.RED, style=Style.NORMAL, back=Back.RESET),
    WARNING=Color(fore=Fore.YELLOW, style=Style.NORMAL, back=Back.RESET),
    INFO=Color(fore=Fore.WHITE, style=Style.NORMAL, back=Back.RESET),
    DEBUG=Color(fore=Fore.GREEN, style=Style.NORMAL, back=Back.RESET)
)


def colorit(message, levelname):
    color = COLORS.get(levelname)
    if color:
        return f'{color.fore}{color.style}{color.back}{message}{Style.RESET_ALL}'
    else:
        return message


def info_color(message):
    return f'{Fore.WHITE}{Style.DIM}{Back.RESET}{message}{Style.RESET_ALL}'


class ColorFormater(logging.Formatter):

    def format(self, record: logging.LogRecord):
        s = self.formatTime(record)
        fmt_log = info_color(f'{s} {record.pathname}:{record.lineno}\n') + \
            colorit(f'{record.levelname} {record.msg}', record.levelname)
        return fmt_log


def make_stream_handler():
    color_formater = ColorFormater()
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(color_formater)
    return stream_handler


def make_file_handler(_log_path=None):
    # Set default value
    if not _log_path:
        _log_path = DEFAULT_LOG_PATH

    # Add pid and thread name in log file
    file_formater = logging.Formatter(
        fmt='%(asctime)s [%(levelname)s] [PID]%(process)s::%(threadName)s (%(pathname)s)  %(message)s'
    )
    log_file = Path(_log_path).expanduser().absolute().resolve()

    file_handler = TimedRotatingFileHandler(log_file, backupCount=1, encoding='utf-8', when='midnight')
    file_handler.setFormatter(file_formater)
    return file_handler


def init(config):
    global LOGGER_INITED
    if LOGGER_INITED:
        return

    if not config:
        config = {}

    logging.addLevelName(60, 'NOTICE')

    stream_handler = make_stream_handler()

    log_path = config.get('log')
    file_handler = make_file_handler(log_path)

    verbose = config.get('verbose', 0)
    if verbose == 0:
        logger_level = logging.ERROR
    elif verbose == 1:
        logger_level = logging.INFO
    elif verbose >= 2:
        logger_level = logging.DEBUG

    lyrebird_logger = logging.getLogger('lyrebird')
    lyrebird_logger.addHandler(stream_handler)
    lyrebird_logger.addHandler(file_handler)
    lyrebird_logger.setLevel(logger_level)

    for _logger_name in ['socketio', 'engineio', 'mock', 'werkzeug', 'flask']:
        _logger = logging.getLogger(_logger_name)
        _logger.addHandler(file_handler)
        _logger.setLevel(logger_level)

    LOGGER_INITED = True


def get_logger() -> logging.Logger:
    return logging.getLogger('lyrebird')



from lyrebird.log import get_logger
from .server import serve, publish_init_status
from lyrebird.base_server import ProcessServer

logger = get_logger()


class ExtraMockServer(ProcessServer):

    def run(self, queue, config, *args, **kwargs):
        publish_init_status(queue, 'READY')
        serve(queue, config, *args, **kwargs)



from lyrebird import application
from .function_executor import FunctionExecutor


class FlowEditorHandler(FunctionExecutor):

    def __init__(self):
        self.on_request = application.on_request
        self.on_response = application.on_response
        self.on_request_upstream = application.on_request_upstream
        self.on_response_upstream = application.on_response_upstream

    def on_request_handler(self, handler_context):
        matched_funcs = FlowEditorHandler.get_matched_sorted_handler(self.on_request, handler_context.flow)
        if not matched_funcs:
            return

        self.script_executor(matched_funcs, handler_context.flow)
        handler_context.set_request_edited()
        handler_context.flow['request']['headers']['lyrebird_modified'] = 'modified'

    def on_request_upstream_handler(self, handler_context):
        matched_funcs = FlowEditorHandler.get_matched_sorted_handler(self.on_request_upstream, handler_context.flow)
        if not matched_funcs:
            return

        self.script_executor(matched_funcs, handler_context.flow)
        handler_context.set_request_edited()
        handler_context.flow['request']['headers']['lyrebird_modified'] = 'modified'

    def on_response_upstream_handler(self, handler_context):
        matched_funcs = FlowEditorHandler.get_matched_sorted_handler(self.on_response_upstream, handler_context.flow)
        if not matched_funcs:
            return

        if not handler_context.flow['response'].get('data'):
            handler_context.update_response_data2flow()

        self.script_executor(matched_funcs, handler_context.flow)
        handler_context.set_response_edited()
        handler_context.flow['response']['headers']['lyrebird_modified'] = 'modified'

    def on_response_handler(self, handler_context):
        matched_funcs = FlowEditorHandler.get_matched_sorted_handler(self.on_response, handler_context.flow)
        if not matched_funcs:
            return

        # Set default value for handler function
        if not handler_context.flow['response'].get('code'):
            handler_context.flow['response']['code'] = 200
        if not handler_context.flow['response'].get('headers'):
            handler_context.flow['response']['headers'] = {}
        if not handler_context.flow['response'].get('data'):
            if handler_context.response:
                handler_context.update_response_data2flow()
            else:
                handler_context.flow['response']['data'] = None

        self.script_executor(matched_funcs, handler_context.flow)
        handler_context.set_response_edited()
        handler_context.flow['response']['headers']['lyrebird_modified'] = 'modified'

    @staticmethod
    def script_executor(func_list, flow):
        application.encoders_decoders.decoder_handler(flow)
        FlowEditorHandler.func_handler(func_list, flow)
        application.encoders_decoders.encoder_handler(flow)



from flask_restful import Resource
from lyrebird.mock import context
from lyrebird import application


class SearchMockDataByName(Resource):

    def get(self, search_str=None):
        _matched_group = []

        def _add_group(_group):
            _matched_group.append({
                'id': _group['id'],
                'name': _group['name'],
                'type': _group['type'],
                'parent_id': _group['parent_id'],
                'abs_parent_path': _get_abs_parent_path(_group),
                'abs_parent_obj': _get_abs_parent_obj(_group)
            })

        def _get_abs_parent_path(_group):
            _abs_parent_path = context.application.data_manager._get_abs_parent_path(_group)
            return _abs_parent_path

        def _get_abs_parent_obj(_group):
            _abs_parent_obj = context.application.data_manager._get_abs_parent_obj(_group)
            return _abs_parent_obj

        if search_str:
            if context.application.data_manager.id_map.get(search_str):
                # search_str is group id
                group = context.application.data_manager.id_map.get(search_str)
                _add_group(group)
            else:
                for _id, group in context.application.data_manager.id_map.items():
                    if group.get('type') == 'group' and search_str.lower() in group.get('name', '').lower():
                        _add_group(group)
        else:
            for _id, group in context.application.data_manager.id_map.items():
                if group.get('type') == 'group' and group.get('name'):
                    _add_group(group)
        return application.make_ok_response(data=_matched_group)



import os
import json
import codecs
from pathlib import Path
from urllib.parse import urlparse


PROP_FILE_NAME = '.lyrebird_prop'


class FileDataAdapter:

    def __init__(self, context_self):
        self.context = context_self

    def _set_root(self, root_path):
        """
        Set a new mock data root dir
        -----
        DataManager will reload all data from this new root dir.
        And all activited data will be removed from DataManager.
        """
        _root_path = Path(root_path).expanduser()
        if not _root_path.exists():
            _root_path.mkdir(parents=True, exist_ok=True)
        if not _root_path.is_dir():
            raise RootPathIsNotDir(root_path)
        self.context.root_path = _root_path
        self.context.reload()

    def _reload(self):
        self.context.id_map = {}
        self._load_prop()
        self.context._sort_children_by_name()

    def _load_prop(self):
        path = self.context.root_path / PROP_FILE_NAME
        if not path.exists():
            self._save_prop()
        with codecs.open(path) as f:
            _root_prop = json.load(f)
        self.context.root = _root_prop
        self._init_id_map(self.context.root)

    def _init_id_map(self, node):
        if 'id' in node:
            self.context.id_map[node['id']] = node
        if 'children' in node:
            for child in node['children']:
                self._init_id_map(child)

    def _get_group(self, id_):
        return self.context.id_map.get(id_)

    def _add_group(self, data, **kwargs):
        self._save_prop()

    def _update_group(self, data):
        self._save_prop()

    def _delete_group(self, data):
        self._save_prop()

    def _delete_group_by_query(self, query):
        self._save_prop()

    # data
    def _load_data(self, data_id, path=None):
        if not path:
            path = self.context.root_path / data_id
        if not Path(path).exists():
            raise DataNotFound(f'Data file {path}')
        with codecs.open(path) as f:
            return json.load(f)

    def _load_data_by_query(self, query):
        datas = []
        for _id in query.get('id', []):
            datas.append(self._load_data(_id))
        return datas

    def _add_data(self, data, path=None):
        if not path:
            path = self.context.root_path / data['id']
            self._add_group(data)
        self._save_data(path, data)

    def _delete_data(self, _id):
        data_file_path = self.context.root_path / _id
        os.remove(data_file_path)

    def _delete_data_by_query(self, query):
        for _id in query.get('id', []):
            self._delete_data(_id)

    def _update_data(self, data):
        self._save_data(self.context.root_path/data['id'], data)

    def _save_prop(self):
        # TODO: Handle frequent writes
        self.context._sort_children_by_name()
        prop_writer = PropWriter()
        prop_writer.dict_ignore_key.update(self.context.unsave_keys)

        prop_str = prop_writer.parse(self.context.root)
        prop_file = self.context.root_path / PROP_FILE_NAME
        with codecs.open(prop_file, 'w') as f:
            f.write(prop_str)
        self.context.reactive()

    def _save_data(self, path, data):
        prop_str = json.dumps(data, ensure_ascii=False)
        with codecs.open(path, 'w') as f:
            f.write(prop_str)

    def _get_data_name(self, data):
        name = 'New Data'
        request = data.get('request')
        if not request:
            return name

        url = request.get('url')
        if not url:
            return name

        parsed_url = urlparse(url)
        host = parsed_url.hostname
        path = parsed_url.path
        if path:
            name = path
        elif host:
            name = host
        else:
            name = url[0:100]

        return name

    def _get_data_rule(self, request):
        pattern = 'YOUR-REQUEST-PATH'
        url = request.get('url')
        if not url:
            return {'request.url': f'(?=.*{pattern})'}

        parsed_url = urlparse(url)
        host = parsed_url.netloc
        path = parsed_url.path
        query = parsed_url.query
        if path and host:
            pattern = path
            if query:
                pattern += '\?' 
            elif url.endswith(path):
                pattern += '$'

        elif host:
            pattern = host
            if url.endswith(host):
                pattern += '$'

        else:
            pattern = url
        return {'request.url': f'(?=.*{pattern})'}

    def _get_activate_group(self, search_id):
        return self.context.id_map.get(search_id)

    def _after_activate(self, **kwargs):
        pass

    # duplicate
    def duplicate(self, _id):
        self.context.copy(_id)
        _node = self.context.id_map.get(_id)
        if not _node:
            raise IDNotFound(_id)
        parent_id = _node.get('parent_id')
        if not parent_id:
            raise IDNotFound(parent_id)

        origin_name = _node.get('name')

        new_uuid = self.context.paste(parent_id)
        duplicate_node = self.context.id_map.get(new_uuid)
        duplicate_node_name = duplicate_node.get('name')
        return {
            'message': f'Duplicat group {origin_name} success! new group name: {duplicate_node_name}',
            'id': new_uuid
        }


class PropWriter:

    def __init__(self):
        self.indent = 0
        self.parsers = {
            'dict': self.dict_parser,
            'list': self.list_parser,
            'int': self.int_parser,
            'str': self.str_parser,
            'bool': self.bool_parser,
            'NoneType': self.none_parser
        }
        self.dict_ignore_key = set()

    def parse(self, prop):
        prop_type = type(prop)
        parser = self.parsers.get(prop_type.__name__)
        if not parser:
            raise DumpPropError(f'Not support type {prop_type}')
        return parser(prop)

    def dict_parser(self, val):
        dict_str = '{'
        children = None
        for k, v in val.items():
            if k in self.dict_ignore_key:
                continue
            if k == 'children':
                children = v
                continue
            dict_str += f'"{k}":{self.parse(v)},'
        if children is not None:
            dict_str += self.children_parser(children)
        if dict_str.endswith(','):
            dict_str = dict_str[:-1]
        dict_str += '}'
        return dict_str

    def list_parser(self, val):
        list_str = '['
        for item in val:
            item_str = self.parse(item)
            list_str += item_str + ','
        if list_str.endswith(','):
            list_str = list_str[:-1]
        list_str += ']'
        return list_str

    def int_parser(self, val):
        return f'{val}'

    def str_parser(self, val):
        return json.dumps(val, ensure_ascii=False)

    def bool_parser(self, val):
        return json.dumps(val)

    def none_parser(self, val):
        return 'null'

    def children_parser(self, val):
        self.indent += 1
        children_str = '"children":['
        for child in val:
            child_str = self.parse(child)
            children_str += '\n' + '  '*self.indent + child_str + ','
        if children_str.endswith(','):
            children_str = children_str[:-1]
        children_str += ']'
        self.indent -= 1
        return children_str


class IDNotFound(Exception):
    pass


class DataNotFound(Exception):
    pass


class DumpPropError(Exception):
    pass


class RootPathIsNotDir(Exception):
    pass


data_adapter = FileDataAdapter



from aiohttp import web, client
from typing import List, Set, Optional
from urllib import parse as urlparse
import re
from lyrebird import utils


class UnknownLyrebirdProxyProtocol(Exception):
    pass


class LyrebirdProxyContext:
    def __init__(self):
        self.request: web.Request = None
        self.netloc = None
        self.origin_url = None
        self.forward_url = None
        self.init = False
        self.protocol_parsers = [
            self.protocol_read_from_path,
            self.protocol_read_from_header,
            self.protocol_read_from_query,
            self.protocol_read_from_query_2
        ]

    def protocol_read_from_path(self, request: web.Request, lb_config):
        # Lyrebird proxy protocol #1
        # Origin target info in path
        # e.g.
        # http://{lyrebird_host}/{origin_full_url}

        if not request.path.startswith('/http://') and not request.path.startswith('/https://'):
            return

        origin_full_url = request.path_qs[1:]
        url = urlparse.urlparse(origin_full_url)

        # origin url for proxy
        self.origin_url = origin_full_url
        # forward url to lyrebird main port 'default 9090'
        port = lb_config.get('mock.port')
        self.forward_url = f'http://127.0.0.1:{port}/mock/{origin_full_url}'

        self.netloc = url.netloc
        self.request = request

        # Set init success
        self.init = True

    def protocol_read_from_header(self, request: web.Request, lb_config):
        # Lyrebird proxy protocol #2
        # Origin target info in headers
        # e.g.
        # http://{lyrebird_host}/{origing_path_and_query}
        # headers:
        # MKScheme: {origin_scheme}
        # MKOriginHost: {origin_host}
        # MKOriginPort: {origin_port}
        header_name_scheme = lb_config.get('mock.proxy_headers', {}).get('scheme')
        header_name_host = lb_config.get('mock.proxy_headers', {}).get('host')
        header_name_port = lb_config.get('mock.proxy_headers', {}).get('port')

        if not request.headers.get(header_name_host) or not request.headers.get(header_name_scheme):
            return

        target_scheme = request.headers.get(header_name_scheme)
        target_host = request.headers.get(header_name_host)
        target_port = request.headers.get(header_name_port)
        if target_port:
            origin_full_url = f'{target_scheme}://{target_host}:{target_port}{request.rel_url}'
            netloc = f'{target_host}:{target_port}'
        else:
            origin_full_url = f'{target_scheme}://{target_host}{request.rel_url}'
            netloc = target_host

        # origin url for proxy
        self.origin_url = origin_full_url
        # forward url to lyrebird main port 'default 9090'
        port = lb_config.get('mock.port')
        self.forward_url = f'http://127.0.0.1:{port}/mock/{origin_full_url}'
        self.netloc = netloc
        self.request = request

        # Set init success
        self.init = True

    def protocol_read_from_query(self, request: web.Request, lb_config):
        # Lyrebird proxy protocol #3
        # Origin target info in query
        # key is "proxy"
        # e.g.
        # http://{lyrebird_host}/{origing_path}?{proxy=encoded-url}

        # default key=proxy
        if not request.query.get('proxy'):
            return

        query_list = str(request.url).split('proxy=')
        origin_url = query_list[1] if len(query_list) > 1 else ''

        unquote_origin_url = urlparse.unquote(origin_url)
        url_obj = urlparse.urlparse(unquote_origin_url)

        self.origin_url = unquote_origin_url
        # forward url to lyrebird main port 'default 9090'
        port = lb_config.get('mock.port')
        self.forward_url = f'http://127.0.0.1:{port}/mock/?proxy={origin_url}'
        self.netloc = url_obj.netloc
        self.request = request

        # Set init success
        self.init = True

    def protocol_read_from_query_2(self, request: web.Request, lb_config):
        # Lyrebird proxy protocol #4
        # Origin target info in query
        # Set proxyscheme、proxyhost、proxypath
        # e.g.
        # http://{lyrebird_host}?proxyscheme={origin_scheme}&proxyhost={urlencode(origin_host)}&proxypath={urlencode(origin_path)}&{origin_query}
        proxy_scheme = request.query.get('proxyscheme', 'http')
        proxy_host = request.query.get('proxyhost', None)
        proxy_path = request.query.get('proxypath', '/')

        if not proxy_host:
            return

        origin_query_str = ''
        raw_query_array = utils.get_query_array(request.path_qs)
        if raw_query_array:
            # query string to 2D array
            # like a=1&b=2 ==> [(a, 1), (b, 2)]
            raw_query_items = list(zip(raw_query_array[::2], raw_query_array[1::2]))
            # remove lyrebrid proxy protocol keys from query string
            raw_query_items = list(filter(lambda x: x[0] not in ['proxyscheme',
                                   'proxyhost', 'proxypath'], raw_query_items))
            # 2D array to query string
            origin_query_str = '?'+'&'.join([f'{item[0]}={item[1]}' for item in raw_query_items])

        origin_url = f'{proxy_scheme}://{urlparse.unquote(proxy_host)}{urlparse.unquote(proxy_path)}{origin_query_str}'

        self.origin_url = origin_url
        url = urlparse.urlparse(origin_url)

        # forward url to lyrebird main port 'default 9090'
        port = lb_config.get('mock.port')
        self.forward_url = f'http://127.0.0.1:{port}/mock/{origin_url}'
        self.netloc = url.netloc
        self.request = request

        # Set init success
        self.init = True

    @classmethod
    def parse(cls, request: web.Request, lb_config):
        ctx = cls()
        for parser in ctx.protocol_parsers:
            parser(request, lb_config)
            if ctx.init:
                return ctx
        raise UnknownLyrebirdProxyProtocol



import uuid
import time
import ipaddress
import json

from .. import context
from lyrebird import utils
from lyrebird import application
from lyrebird.log import get_logger
from lyrebird.mock.blueprints.apis.bandwidth import config
from urllib.parse import urlparse, unquote
from .http_data_helper import DataHelper
from .http_header_helper import HeadersHelper


logger = get_logger()


class HandlerContext:
    """
    请求处理器上下文变量
    用于保存一个请求处理过程中的request, response

    """
    MOCK_PATH_PREFIX = '/mock'

    def __init__(self, request):
        self.id = str(uuid.uuid4())
        self.request = request
        self.response = None
        self.client_req_time = None
        self.client_resp_time = None
        self.server_req_time = None
        self.server_resp_time = None
        self.flow = utils.HookedDict({
            'id': self.id,
            'size': 0,
            'duration': 0,
            'start_time': time.time(),
            'request': utils.HookedDict({
                'headers': utils.CaseInsensitiveDict({})
            }),
            'response': utils.HookedDict({
                'headers': utils.CaseInsensitiveDict({})
            })
        })
        self.request_chain = []
        self.response_chain = []
        self.client_address = None
        self.is_request_edited = False
        self.is_response_edited = False
        self.response_source = ''
        self.is_proxiable = True
        self.response_chunk_size = 2048
        self._parse_request()

    def _parse_request(self):
        # Read stream
        self.request.get_data()
        
        raw_headers = None
        # Read raw headers
        if 'Proxy-Raw-Headers' in self.request.headers:
            raw_headers = json.loads(self.request.headers['Proxy-Raw-Headers'])

        # parse path
        request_info = self._read_origin_request_info_from_url()
        if not request_info['host']:
            request_info_from_header = self._read_origin_request_info_from_header(headers=raw_headers)
            if len(request_info_from_header) > 0:
                request_info = request_info_from_header

        # parse query
        query_array = utils.get_query_array(request_info['url'])
        query = dict(zip(query_array[::2], query_array[1::2]))
        request_info['query'] = query

        if raw_headers:
            headers = raw_headers
        else:
            headers = HeadersHelper.origin2flow(self.request)

        _request = dict(
            headers=headers,
            method=self.request.method,
            timestamp=round(time.time(), 3)
        )
        _request.update(request_info)

        # handle request data
        if self.request.method in ['POST', 'PUT']:
            DataHelper.origin2flow(self.request, output=_request, chain=self.request_chain)

        if self.request.headers.get('Lyrebird-Client-Address'):
            self.client_address = self.request.headers.get('Lyrebird-Client-Address')
        else:
            self.client_address = self.request.remote_addr
        self.flow['client_address'] = self.client_address

        self.flow['request'] = _request
        context.application.cache.add(self.flow)

        logger.debug(f'[On client request] {self.flow["request"]["url"]}')

    def _read_origin_request_info_from_url(self):
        url_prefix = '/'+self.request.blueprint+'/'
        raw_url = self.request.path[len(url_prefix):]
        if self.request.query_string:
            raw_url += '?' + self.request.query_string.decode()
        return self._get_parse_url_dict(raw_url)

    @staticmethod
    def _get_parse_url_dict(raw_url):
        parsed_path = urlparse(raw_url)
        # urllib.unquote : fix bug - url contains ',' will be auto encoded by flask, that cause proxy not work.
        # e.g /1.2,3 -> 1.2%2C3
        _request = dict(
            url=raw_url,
            scheme=parsed_path.scheme,
            host=parsed_path.hostname,
            port=parsed_path.port if parsed_path.port else '80',
            path=unquote(parsed_path.path)
        )
        return _request

    def _read_origin_request_info_from_header(self, headers=None):
        proxy_headers = application.config['mock.proxy_headers']
        _headers = headers if headers else self.request.headers
        scheme = _headers.get(proxy_headers['scheme'], 'http')
        host = _headers.get(proxy_headers['host'])
        port = _headers.get(proxy_headers['port'], None)
        if scheme == 'http' and not port:
            port = '80'
        elif scheme == 'https' and not port:
            port = '443'

        if not host:
            return {}
            
        scheme = scheme.strip()
        host = host.strip()
        port = port.strip()
        # if host is IP address then full_host=host:port
        # else if is a domain the full_host=host
        full_host = host
        try:
            ipaddress.ip_address(host)
            full_host = host + ':' + port
        except Exception:
            pass
        return dict(
            url=scheme+'://'+full_host+self.request.full_path[len(self.MOCK_PATH_PREFIX):],
            scheme=scheme,
            host=host,
            port=port,
            path=self.request.path[len(self.MOCK_PATH_PREFIX):]
        )

    def set_request_edited(self):
        self.is_request_edited = True

    def set_response_edited(self):
        self.is_response_edited = True

    def set_response_source_mock(self):
        self.response_source = 'mock'

    def set_response_source_proxy(self):
        self.response_source = 'proxy'

    def get_request_body(self):
        if self.is_request_edited:
            # TODO Repeated calls, remove it
            self.flow['request']['headers'] = HeadersHelper.flow2origin(self.flow['request'], chain=self.request_chain)

            _data = DataHelper.flow2origin(self.flow['request'], chain=self.request_chain)
        else:
            _data = self.request.data or self.request.form or None
        return _data

    def get_request_headers(self):
        if self.is_request_edited:
            self.flow['request']['headers'] = HeadersHelper.flow2origin(self.flow['request'], chain=self.request_chain)

        headers = {}
        unproxy_headers = application.config.get('proxy.ignored_headers', {})
        for name, value in self.flow['request']['headers'].items():
            if not value or name in ['Cache-Control', 'Host', 'Transfer-Encoding']:
                continue
            if name in unproxy_headers and unproxy_headers[name] in value:
                continue
            headers[name] = value
        return headers

    def get_response_generator(self):
        if self.is_response_edited:
            self.flow['response']['headers'] = HeadersHelper.flow2origin(self.flow['response'], chain=self.response_chain)
            _generator = self._generator_bytes()
        else:
            _generator = self._generator_stream()
        return _generator

    def _generator_bytes(self):
        def generator():
            try:
                _resp_data = DataHelper.flow2origin(self.flow['response'], chain=self.response_chain) or ''
                length = len(_resp_data)
                size = self.response_chunk_size
                bandwidth = config.bandwidth
                if bandwidth > 0:
                    sleep_time = self.response_chunk_size / (bandwidth * 1024)
                else:
                    sleep_time = 0
                for i in range(int(length/size) + 1):
                    time.sleep(sleep_time)
                    self.server_resp_time = time.time()
                    yield _resp_data[ i * size : (i+1) * size ]
            finally:
                self.update_client_resp_time()
        return generator

    def _generator_stream(self):
        def generator():
            upstream = self.response
            try:
                bandwidth = config.bandwidth
                if bandwidth > 0:
                    sleep_time = self.response_chunk_size / (bandwidth * 1024)
                else:
                    sleep_time = 0
                buffer = []
                for item in upstream.response:
                    buffer.append(item)
                    time.sleep(sleep_time)
                    self.server_resp_time = time.time()
                    yield item
            finally:
                self.response.data = b''.join(buffer)
                DataHelper.origin2flow(self.response, output=self.flow['response'], chain=self.response_chain)

                self.update_client_resp_time()
                upstream.close()
        return generator

    def update_response_headers_code2flow(self, output_key='response'):
        self.flow[output_key]  = {
            'code': self.response.status_code,
            'timestamp': round(time.time(), 3)
        }
        HeadersHelper.origin2flow(self.response, output=self.flow[output_key])

    def update_response_data2flow(self, output_key='response'):
        DataHelper.origin2flow(self.response, output=self.flow[output_key], chain=self.response_chain)

    def update_client_req_time(self):
        self.client_req_time = time.time()
        # 消息总线 客户端请求事件，启用此事件
        method = self.flow['request']['method']
        url = self.flow['request']['url']

        _flow_client_req = {}
        for key, value in self.flow.items():
            _flow_client_req[key] = value

        context.application.event_bus.publish(
            'flow.request',
            dict(
                flow=_flow_client_req,
                message=f"URL: {url}\nMethod: {method}\n"
            )
        )

    def update_client_resp_time(self):
        self.client_resp_time = time.time()
        # 消息总线 客户端响应事件，启用此事件
        resp_data = self.flow['response'].get('data', '')
        if isinstance(resp_data, str):
            self.flow['size'] = len(resp_data.encode())
        else:
            self.flow['size'] = len(resp_data)

        self.flow['duration'] = self.server_resp_time - self.client_req_time

        method = self.flow['request']['method']
        url = self.flow['request']['url']
        code = self.flow['response']['code']
        duration = utils.convert_time(self.flow['duration'])
        size = utils.convert_size(self.flow['size'])

        parsed_url = self._get_parse_url_dict(url)
        self.flow['request'].update(parsed_url)

        # Import decoder for decoding the requested content
        decode_flow = {}
        application.encoders_decoders.decoder_handler(self.flow, output=decode_flow)

        context.application.event_bus.publish(
            'flow',
            dict(
                flow=decode_flow,
                message=f"URL: {url}\nMethod: {method}\nStatusCode: {code}\nDuration: {duration}\nSize: {size}",
                export=dict(converter='flow_json')
            )
        )

        if context.application.work_mode == context.Mode.RECORD:
            dm = context.application.data_manager
            dm.save_data(self.flow)

    def update_server_req_time(self):
        self.server_req_time = time.time()
        # 消息总线 向服务端转发请求事件，暂不使用
        # context.application.event_bus.publish('flow',
        #                                       dict(name='server.request',
        #                                       time=self.server_req_time,
        #                                       id=self.id,
        #                                       flow=self.flow))

    def update_server_resp_time(self):
        self.server_resp_time = time.time()
        # 消息总线 服务端响应请求事件，暂不使用
        # context.application.event_bus.publish('flow',
        #                                       dict(name='server.response',
        #                                       time=self.server_resp_time,
        #                                       id=self.id,
        #                                       flow=self.flow))

    def add_flow_action(self, action):
        if self.flow.get('action'):
            self.flow['action'].append(action)
        else:
            self.flow['action'] = [action]



from flask_restful import Resource
from lyrebird import application


class StatusBar(Resource):

    def get(self, item_id=None):
        if item_id == None:
            return self._get_all_status_item()
        else:
            return self._get_menu_by_status_item_name(item_id)

    def _get_all_status_item(self):
        status_dict = {}

        plugin_server = application.server['plugin']
        for plugin in plugin_server.plugins.values():
            for status_item in plugin.status:
                if status_item.placement not in status_dict:
                    status_dict[status_item.placement] = []
                status_detail = status_item.json()
                status_dict[status_item.placement].append(status_detail)

        for status_list in status_dict.values():
            status_list.sort(key=lambda x:x['rank'], reverse=True)
        return application.make_ok_response(**status_dict)

    def _get_menu_by_status_item_name(self, name):
        plugin_server = application.server['plugin']
        for plugin in plugin_server.plugins.values():
            for status_item in plugin.status:
                if status_item.name != name:
                    continue

                menu_list = status_item.get_detail()
                return application.make_ok_response(data=menu_list)

        return application.make_fail_response(f'Status item not found, id={name}')



from pathlib import Path
from lyrebird import log
import subprocess
import os
import json
import requests
import time
from lyrebird.base_server import ProcessServer
from lyrebird.mitm.mitm_installer import init_mitm

"""
HTTP proxy server
Default port 4272
"""


class LyrebirdProxyServer(ProcessServer):

    def __init__(self):
        super().__init__()
        self.mitm_path = init_mitm()
        self.kwargs['mitm_path'] = self.mitm_path

    def show_mitmdump_start_timeout_help(self, mitmdump_filepath, logger):
        logger.error(f'Start mitmdump failed.\nPlease check your mitmdump file {mitmdump_filepath}')

    def wait_for_mitm_start(self, config, logger):
        timeout = 30
        wait_time_count = 0
        mock_port = config.get('mock.port')
        proxy_port = config.get('proxy.port')
        while True:
            if wait_time_count >= timeout:
                return False

            time.sleep(1)
            wait_time_count += 1
            try:
                resp = requests.get(f'http://127.0.0.1:{mock_port}/api/status',
                                    proxies={'http': f'http://127.0.0.1:{proxy_port}'})
                if resp.status_code != 200:
                    continue
                else:
                    return True
            except Exception:
                continue

    def start_mitmdump(self, queue, config, logger, mitmdump_path):
        proxy_port = config.get('proxy.port', 4272)
        mock_port = config.get('mock.port', 9090)
        '''
        --ignore_hosts:
        The ignore_hosts option allows you to specify a regex which is matched against a host:port
        string (e.g. “example.com:443”) of a connection. Matching hosts are excluded from interception,
        and passed on unmodified.

        # Ignore everything but sankuai.com, meituan.com and dianping.com:
        --ignore-hosts '^(?!.*sankuai.*)(?!.*meituan.*)(?!.*dianping.*)'

        According to mitmproxy docs: https://docs.mitmproxy.org/stable/howto-ignoredomains/
        '''
        ignore_hosts = config.get('proxy.ignore_hosts', None)

        current_path = Path(__file__).parent
        script_path = current_path/'mitm_script.py'

        mitm_arguments = [
            '-s', str(script_path),
            '-p', str(proxy_port),
            '--ssl-insecure',
            '--no-http2',
            '-q',
            '--set',
            'block_global=false'
        ]
        if ignore_hosts:
            mitm_arguments += ['--ignore-hosts', ignore_hosts]
        mitmenv = os.environ
        mitmenv['PROXY_PORT'] = str(mock_port)
        mitmenv['PROXY_FILTERS'] = json.dumps(config.get('proxy.filters', []))
        logger.info('HTTP proxy server starting...')
        subprocess.Popen(f'{str(mitmdump_path)} {" ".join(mitm_arguments)}', shell=True, env=mitmenv)
        is_mitm_start = self.wait_for_mitm_start(config, logger)
        if is_mitm_start:
            self.publish_init_status(queue, 'READY')
            logger.log(60, f'HTTP proxy server start on {proxy_port}')
        else:
            self.publish_init_status(queue, 'ERROR')
            self.show_mitmdump_start_timeout_help(mitmdump_path, logger)

    def publish_init_status(self, queue, status):
        queue.put({
            'type': 'event',
            "channel": "system",
            "content": {
                'system': {
                    'action': 'init_module',
                    'status': status,
                    'module': 'mitm_proxy'
                }
            }
        })

    def run(self, queue, config, *args, **kwargs):
        # Init logger
        log.init(config)
        logger = log.get_logger()
        mitm_path = kwargs.get('mitm_path')
        self.start_mitmdump(queue, config, logger, mitm_path)


class UnsupportedPlatform(Exception):
    pass
